@@ -1,12 +1,12 @@
-(***********************************************************************)
-(*  v      *   The Coq Proof Assistant  /  The Coq Development Team    *)
-(* <O___,, *        INRIA-Rocquencourt  &  LRI-CNRS-Orsay              *)
-(*   \VV/  *************************************************************)
-(*    //   *      This file is distributed under the terms of the      *)
-(*         *       GNU Lesser General Public License Version 2.1       *)
-(***********************************************************************)
+(************************************************************************)
+(*  v      *   The Coq Proof Assistant  /  The Coq Development Team     *)
+(* <O___,, * CNRS-Ecole Polytechnique-INRIA Futurs-Universite Paris Sud *)
+(*   \VV/  **************************************************************)
+(*    //   *      This file is distributed under the terms of the       *)
+(*         *       GNU Lesser General Public License Version 2.1        *)
+(************************************************************************)
 
-(* $Id: termops.mli,v 1.8 2002/03/26 16:29:59 mohring Exp $ *)
+(*i $Id: termops.mli 9314 2006-10-29 20:11:08Z herbelin $ i*)
 
 open Util
 open Pp
@@ -16,13 +16,24 @@ open Sign
 open Environ
 
 (* Universes *)
-val set_module : Names.dir_path -> unit
 val new_univ : unit -> Univ.universe
 val new_sort_in_family : sorts_family -> sorts
+val new_Type : unit -> types
+val new_Type_sort : unit -> sorts
+val refresh_universes : types -> types
+
+(* printers *)
+val print_sort : sorts -> std_ppcmds
+val pr_sort_family : sorts_family -> std_ppcmds
+(* debug printer: do not use to display terms to the casual user... *)
+val set_print_constr : (env -> constr -> std_ppcmds) -> unit
+val print_constr     : constr -> std_ppcmds
+val print_constr_env : env -> constr -> std_ppcmds
+val print_named_context : env -> std_ppcmds
+val print_rel_context : env -> std_ppcmds
+val print_env : env -> std_ppcmds
 
 (* iterators on terms *)
-val print_sort : sorts -> std_ppcmds
-val print_sort_family : sorts_family -> std_ppcmds
 val prod_it : init:types -> (name * types) list -> types
 val lam_it : init:constr -> (name * types) list -> constr
 val rel_vect : int -> int -> constr array
@@ -42,6 +53,10 @@ val it_named_context_quantifier :
   (named_declaration -> 'a -> 'a) -> init:'a -> named_context -> 'a
 val it_mkNamedProd_or_LetIn   : init:types -> named_context -> types
 val it_mkNamedLambda_or_LetIn : init:constr -> named_context -> constr
+val it_mkNamedProd_wo_LetIn   : init:types -> named_context -> types
+
+(**********************************************************************)
+(* Generic iterators on constr                                        *)
 
 val map_constr_with_named_binders :
   (name -> 'a -> 'a) ->
@@ -54,6 +69,22 @@ val map_constr_with_full_binders :
   (rel_declaration -> 'a -> 'a) ->
   ('a -> constr -> constr) -> 'a -> constr -> constr
 
+(* [fold_constr_with_binders g f n acc c] folds [f n] on the immediate
+   subterms of [c] starting from [acc] and proceeding from left to
+   right according to the usual representation of the constructions as
+   [fold_constr] but it carries an extra data [n] (typically a lift
+   index) which is processed by [g] (which typically add 1 to [n]) at
+   each binder traversal; it is not recursive *)
+
+val fold_constr_with_binders :
+  ('a -> 'a) -> ('a -> 'b -> constr -> 'b) -> 'a -> 'b -> constr -> 'b
+
+val iter_constr_with_full_binders :
+  (rel_declaration -> 'a -> 'a) -> ('a -> constr -> unit) -> 'a -> 
+    constr -> unit
+
+(**********************************************************************)
+
 val strip_head_cast : constr -> constr
 
 (* occur checks *)
@@ -71,48 +102,51 @@ val occur_term : constr -> constr -> bool
 val free_rels : constr -> Intset.t
 
 (* Substitution of metavariables *)
-val subst_meta : (int * constr) list -> constr -> constr
-
-(* Expansion of local definitions *)
-val whd_locals : env -> constr -> constr
-val nf_locals  : env -> constr -> constr
+type metamap = (metavariable * constr) list 
+val subst_meta : metamap -> constr -> constr
 
 (* [pop c] lifts by -1 the positive indexes in [c] *)
 val pop : constr -> constr
 
-(* substitution of an arbitrary large term. Uses equality modulo
+(* bindings of an arbitrary large term. Uses equality modulo
    reduction of let *)
 val dependent : constr -> constr -> bool
 val subst_term_gen :
-  (env -> constr -> constr -> bool) -> env -> constr -> constr -> constr
+  (constr -> constr -> bool) -> constr -> constr -> constr
 val replace_term_gen :
-  (env -> constr -> constr -> bool) ->
-    env -> constr -> constr -> constr -> constr
+  (constr -> constr -> bool) ->
+    constr -> constr -> constr -> constr
 val subst_term : constr -> constr -> constr
 val replace_term : constr -> constr -> constr -> constr
 val subst_term_occ_gen :
-  env -> int list -> int -> constr -> types -> int * types
-val subst_term_occ : env -> int list -> constr -> types -> types
+  int list -> int -> constr -> types -> int * types
+val subst_term_occ : int list -> constr -> types -> types
 val subst_term_occ_decl :
-  env -> int list -> constr -> named_declaration -> named_declaration
+  int list -> constr -> named_declaration -> named_declaration
+
+val error_invalid_occurrence : int list -> 'a
 
 (* Alternative term equalities *)
-val zeta_eq_constr : constr -> constr -> bool
 val eta_reduce_head : constr -> constr
 val eta_eq_constr : constr -> constr -> bool
 
 (* finding "intuitive" names to hypotheses *)
 val first_char : identifier -> string
 val lowercase_first_char : identifier -> string
-val id_of_global : env -> Nametab.global_reference -> identifier
 val sort_hdchar : sorts -> string
 val hdchar : env -> types -> string
 val id_of_name_using_hdchar :
   env -> types -> name -> identifier
 val named_hd : env -> types -> name -> name
 val named_hd_type : env -> types -> name -> name
-val prod_name : env -> name * types * types -> constr
+
+val mkProd_name : env -> name * types * types -> types
+val mkLambda_name : env -> name * types * constr -> constr
+
+(* Deprecated synonyms of [mkProd_name] and [mkLambda_name] *)
+val prod_name : env -> name * types * types -> types
 val lambda_name : env -> name * types * constr -> constr
+
 val prod_create : env -> types * types -> constr
 val lambda_create : env -> types * constr -> constr
 val name_assumption : env -> rel_declaration -> rel_declaration
@@ -134,19 +168,24 @@ val ids_of_named_context : named_context -> identifier list
 val ids_of_context : env -> identifier list
 val names_of_rel_context : env -> names_context
 
+(* Set of local names *)
+val vars_of_env: env -> Idset.t
+val add_vname : Idset.t -> name -> Idset.t
+
 (* sets of free identifiers *)
 type used_idents = identifier list
 val occur_rel : int -> name list -> identifier -> bool
-val occur_id : env -> name list -> identifier -> constr -> bool
+val occur_id : name list -> identifier -> constr -> bool
 
+val next_global_ident_away : 
+  (*allow section vars:*) bool -> identifier -> identifier list -> identifier
 val next_name_not_occuring :
-  env -> name -> identifier list -> name list -> constr -> identifier
+  bool -> name -> identifier list -> name list -> constr -> identifier
 val concrete_name :
-  env -> identifier list -> name list -> name ->
-  constr -> identifier option * identifier list
+  bool -> identifier list -> name list -> name -> constr -> 
+    name * identifier list
 val concrete_let_name :
-  env -> identifier list -> name list ->
-  name -> constr -> name * identifier list
+  bool -> identifier list -> name list -> name -> constr -> name * identifier list
 val rename_bound_var : env -> identifier list -> types -> types
 
 (* other signature iterators *)
@@ -161,3 +200,14 @@ val make_all_name_different : env -> env
 
 val global_vars : env -> constr -> identifier list
 val global_vars_set_of_decl : env -> named_declaration -> Idset.t
+
+(* Test if an identifier is the basename of a global reference *)
+val is_section_variable : identifier -> bool
+
+val isGlobalRef : constr -> bool
+
+(* Combinators on judgments *)
+
+val on_judgment       : (types -> types) -> unsafe_judgment -> unsafe_judgment
+val on_judgment_value : (types -> types) -> unsafe_judgment -> unsafe_judgment
+val on_judgment_type  : (types -> types) -> unsafe_judgment -> unsafe_judgment
