@@ -1,12 +1,12 @@
-(***********************************************************************)
-(*  v      *   The Coq Proof Assistant  /  The Coq Development Team    *)
-(* <O___,, *        INRIA-Rocquencourt  &  LRI-CNRS-Orsay              *)
-(*   \VV/  *************************************************************)
-(*    //   *      This file is distributed under the terms of the      *)
-(*         *       GNU Lesser General Public License Version 2.1       *)
-(***********************************************************************)
+(************************************************************************)
+(*  v      *   The Coq Proof Assistant  /  The Coq Development Team     *)
+(* <O___,, * CNRS-Ecole Polytechnique-INRIA Futurs-Universite Paris Sud *)
+(*   \VV/  **************************************************************)
+(*    //   *      This file is distributed under the terms of the       *)
+(*         *       GNU Lesser General Public License Version 2.1        *)
+(************************************************************************)
 
-(* $Id: elim.ml,v 1.27 2002/02/15 18:02:04 barras Exp $ *)
+(* $Id: elim.ml 7538 2005-11-08 17:14:52Z herbelin $ *)
 
 open Pp
 open Util
@@ -14,6 +14,7 @@ open Names
 open Term
 open Termops
 open Environ
+open Libnames
 open Reduction
 open Inductiveops
 open Proof_type
@@ -23,6 +24,8 @@ open Tacmach
 open Tacticals
 open Tactics
 open Hiddentac
+open Genarg
+open Tacexpr
 
 let introElimAssumsThen tac ba =
   let nassums = 
@@ -35,13 +38,20 @@ let introElimAssumsThen tac ba =
 
 let introCaseAssumsThen tac ba =
   let case_thin_sign =
-    List.flatten 
-      (List.map 
-	 (function b -> if b then [false;true] else [false]) 
-	 ba.branchsign) 
+    List.flatten
+      (List.map (function b -> if b then [false;true] else [false])
+	ba.branchsign)
   in 
-  let introCaseAssums = intros_clearing case_thin_sign in
-  (tclTHEN introCaseAssums (case_on_ba tac ba))
+  let n1 = List.length case_thin_sign in
+  let n2 = List.length ba.branchnames in
+  let (l1,l2),l3 =
+    if n1 < n2 then list_chop n1 ba.branchnames, []
+    else 
+      (ba.branchnames, []),
+       if n1 > n2 then snd (list_chop n2 case_thin_sign) else [] in
+  let introCaseAssums = tclTHEN (intros_pattern None l1) (intros_clearing l3)
+  in
+  (tclTHEN introCaseAssums (case_on_ba (tac l2) ba))
 
 (* The following tactic Decompose repeatedly applies the
    elimination(s) rule(s) of the types satisfying the predicate
@@ -85,12 +95,12 @@ let up_to_delta = ref false (* true *)
 
 let general_decompose recognizer c gl = 
   let typc = pf_type_of gl c in  
-  (tclTHENS (cut typc) 
-    [tclTHEN (intro_using tmphyp_name)
-      (onLastHyp
-	(ifOnHyp recognizer (general_decompose recognizer)
-          (fun id -> clear [id])));
-     exact_no_check c]) gl
+  tclTHENSV (cut typc) 
+    [| tclTHEN (intro_using tmphyp_name)
+         (onLastHyp
+	    (ifOnHyp recognizer (general_decompose recognizer)
+	      (fun id -> clear [id])));
+       exact_no_check c |] gl
 
 let head_in gls indl t =
   try
@@ -101,19 +111,14 @@ let head_in gls indl t =
     in List.mem ity indl
   with Not_found -> false
        
-let inductive_of_qualid gls qid =
-  let c = 
-    try Declare.construct_qualified_reference qid
-    with Not_found -> Nametab.error_global_not_found qid
-  in
-  match kind_of_term c with
-    | Ind ity -> ity
-    | _ ->
-	errorlabstrm "Decompose"
-	  (Nametab.pr_qualid qid ++ str " is not an inductive type")
+let inductive_of = function
+  | IndRef ity -> ity
+  | r ->
+      errorlabstrm "Decompose"
+        (Printer.pr_global r ++ str " is not an inductive type")
 
 let decompose_these c l gls =
-  let indl = List.map (inductive_of_qualid gls) l in 
+  let indl = (*List.map inductive_of*) l in 
   general_decompose (fun (_,t) -> head_in gls indl t) c gls
 
 let decompose_nonrec c gls =
@@ -131,27 +136,14 @@ let decompose_or c gls =
     (fun (_,t) -> is_disjunction t)
     c gls
 
-let dyn_decompose args gl =
-  let out_qualid = function
-    | Qualid qid -> qid
-    | l -> bad_tactic_args "DecomposeThese" [l] gl in
-  match args with 
-    | Command c :: ids -> 
-        decompose_these (pf_interp_constr gl c) (List.map out_qualid ids) gl
-    | Constr c :: ids  -> 
-	decompose_these c (List.map out_qualid ids) gl
-    | l -> bad_tactic_args "DecomposeThese" l gl
-	  
-let h_decompose =
-  let v_decompose = hide_tactic "DecomposeThese" dyn_decompose in 
-  fun ids c ->
-    v_decompose 
-      (Constr c :: List.map (fun x -> Qualid (Nametab.qualid_of_sp x)) ids)
+let h_decompose l c =
+  Refiner.abstract_tactic (TacDecompose (l,c)) (decompose_these c l)
 
-let vernac_decompose_and = 
-  hide_constr_tactic "DecomposeAnd" decompose_and
-let vernac_decompose_or  = 
-  hide_constr_tactic  "DecomposeOr"  decompose_or
+let h_decompose_or c =
+  Refiner.abstract_tactic (TacDecomposeOr c) (decompose_or c)
+
+let h_decompose_and c =
+  Refiner.abstract_tactic (TacDecomposeAnd c) (decompose_and c)
 
 (* The tactic Double performs a double induction *)
 
@@ -178,11 +170,17 @@ let induction_trailer abs_i abs_j bargs =
 	  in
           let ids = List.rev (ids_of_named_context hyps) in
 	  (tclTHENSEQ
-            [bring_hyps hyps; clear ids; simple_elimination (mkVar id)])
+            [bring_hyps hyps; tclTRY (clear ids); 
+	     simple_elimination (mkVar id)])
           gls))
 
-let double_ind abs_i abs_j gls =
-  let cl = pf_concl gls in 
+let double_ind h1 h2 gls =
+  let abs_i = depth_of_quantified_hypothesis true h1 gls in
+  let abs_j = depth_of_quantified_hypothesis true h2 gls in
+  let (abs_i,abs_j) =
+    if abs_i < abs_j then (abs_i,abs_j) else
+    if abs_i > abs_j then (abs_j,abs_i) else
+      error "Both hypotheses are the same" in
   (tclTHEN (tclDO abs_i intro)
      (onLastHyp
        	(fun id ->
@@ -190,38 +188,7 @@ let double_ind abs_i abs_j gls =
              (introElimAssumsThen (induction_trailer abs_i abs_j))
              ([],[]) (mkVar id)))) gls
 
-let dyn_double_ind = function
-  | [Integer i; Integer j] -> double_ind i j
-  | _ -> assert false
-
-let _ = add_tactic "DoubleInd" dyn_double_ind
+let h_double_induction h1 h2 =
+  Refiner.abstract_tactic (TacDoubleInduction (h1,h2)) (double_ind h1 h2)
 
 
-(*****************************)
-(* Decomposing introductions *)
-(*****************************)
-
-let rec intro_pattern p = 
-  let clear_last = tclLAST_HYP (fun c -> (clear [destVar c]))
-  and case_last  = tclLAST_HYP h_simplest_case in  
-  match p with
-    | WildPat    -> tclTHEN intro clear_last
-    | IdPat  id  -> intro_mustbe_force id
-    | DisjPat l  -> tclTHEN introf
-                       (tclTHENS 
-			  (tclTHEN case_last clear_last)
-			  (List.map intro_pattern l))
-    | ConjPat l  ->
-        tclTHENSEQ [introf; case_last; clear_last; intros_pattern l]
-    | ListPat l  ->  intros_pattern l
-
-and intros_pattern l = tclMAP intro_pattern l
-
-let dyn_intro_pattern = function 
-  | []               -> intros
-  | [Intropattern p] -> intro_pattern p
-  | l                -> bad_tactic_args "Elim.dyn_intro_pattern" l
-
-let v_intro_pattern = hide_tactic "Intros" dyn_intro_pattern
-
-let h_intro_pattern p = v_intro_pattern [Intropattern p]
