@@ -1,12 +1,12 @@
-(***********************************************************************)
-(*  v      *   The Coq Proof Assistant  /  The Coq Development Team    *)
-(* <O___,, *        INRIA-Rocquencourt  &  LRI-CNRS-Orsay              *)
-(*   \VV/  *************************************************************)
-(*    //   *      This file is distributed under the terms of the      *)
-(*         *       GNU Lesser General Public License Version 2.1       *)
-(***********************************************************************)
+(************************************************************************)
+(*  v      *   The Coq Proof Assistant  /  The Coq Development Team     *)
+(* <O___,, * CNRS-Ecole Polytechnique-INRIA Futurs-Universite Paris Sud *)
+(*   \VV/  **************************************************************)
+(*    //   *      This file is distributed under the terms of the       *)
+(*         *       GNU Lesser General Public License Version 2.1        *)
+(************************************************************************)
 
-(* $Id: himsg.ml,v 1.69 2002/04/10 16:07:51 herbelin Exp $ *)
+(* $Id: himsg.ml 9528 2007-01-24 09:43:03Z herbelin $ *)
 
 open Pp
 open Util
@@ -21,16 +21,33 @@ open Sign
 open Environ
 open Pretype_errors
 open Type_errors
+open Indrec
 open Reduction
 open Cases
 open Logic
 open Printer
-open Ast
+open Rawterm
+open Evd
 
-let guill s = "\""^s^"\""
+let quote s = h 0 (str "\"" ++ s ++ str "\"")
+
+let pr_lconstr c = quote (pr_lconstr c)
+let pr_lconstr_env e c = quote (pr_lconstr_env e c)
+let pr_lconstr_env_at_top e c = quote (pr_lconstr_env_at_top e c)
+let pr_ljudge_env e c = let v,t = pr_ljudge_env e c in (quote v,quote t)
+
+let nth i =
+  let many = match i mod 10 with 1 -> "st" | 2 -> "nd" | _ -> "th" in
+  int i ++ str many
+
+let pr_db ctx i =
+  try
+    match lookup_rel i ctx with
+        Name id, _, _ -> pr_id id
+      | Anonymous, _, _ -> str"<>"
+  with Not_found -> str"UNBOUND_REL_"++int i
 
 let explain_unbound_rel ctx n =
-  let ctx = make_all_name_different ctx in
   let pe = pr_ne_context_of (str "In environment") ctx in
   str"Unbound reference: " ++ pe ++
   str"The reference " ++ int n ++ str " is free"
@@ -40,90 +57,102 @@ let explain_unbound_var ctx v =
   str"No such section variable or assumption : " ++ var
 
 let explain_not_type ctx j =
-  let ctx = make_all_name_different ctx in
   let pe = pr_ne_context_of (str"In environment") ctx in
-  let pc,pt = prjudge_env ctx j in
+  let pc,pt = pr_ljudge_env ctx j in
   pe ++ str "the term" ++ brk(1,1) ++ pc ++ spc () ++
   str"has type" ++ spc () ++ pt ++ spc () ++ 
   str"which should be Set, Prop or Type."
 
 let explain_bad_assumption ctx j =
-  let ctx = make_all_name_different ctx in
   let pe = pr_ne_context_of (str"In environment") ctx in
-  let pc,pt = prjudge_env ctx j in
+  let pc,pt = pr_ljudge_env ctx j in
   pe ++ str "cannot declare a variable or hypothesis over the term" ++
   brk(1,1) ++ pc ++ spc () ++ str"of type" ++ spc () ++ pt ++ spc () ++
   str "because this term is not a type."
 
 let explain_reference_variables c =
-  let pc = prterm c in
+  let pc = pr_lconstr c in
   str "the constant" ++ spc () ++ pc ++ spc () ++ 
   str "refers to variables which are not in the context"
 
-let explain_elim_arity ctx ind aritylst c pj okinds = 
+let rec pr_disjunction pr = function
+  | [a] -> pr  a
+  | [a;b] -> pr a ++ str " or" ++ spc () ++ pr b
+  | a::l -> pr a ++ str "," ++ spc () ++ pr_disjunction pr l
+  | [] -> assert false
+
+let explain_elim_arity ctx ind sorts c pj okinds = 
+  let ctx = make_all_name_different ctx in
   let pi = pr_inductive ctx ind in
-  let ppar = prlist_with_sep pr_coma (prterm_env ctx) aritylst in
-  let pc = prterm_env ctx c in
-  let pp = prterm_env ctx pj.uj_val in
-  let ppt = prterm_env ctx pj.uj_type in
+  let pc = pr_lconstr_env ctx c in
   let msg = match okinds with
   | Some(kp,ki,explanation) ->
-      let pki = prterm_env ctx ki in
-      let pkp = prterm_env ctx kp in
+      let pki = pr_sort_family ki in
+      let pkp = pr_sort_family kp in
       let explanation =	match explanation with
 	| NonInformativeToInformative ->
-	  "non-informative objects may not construct informative ones."
+          "proofs can be eliminated only to build proofs"
 	| StrongEliminationOnNonSmallType ->
           "strong elimination on non-small inductive types leads to paradoxes."
 	| WrongArity ->
 	  "wrong arity" in
-	(hov 0 
-           (fnl () ++ str "Elimination of an inductive object of sort : " ++
-            pki ++ brk(1,0) ++
-            str "is not allowed on a predicate in sort : " ++ pkp  ++fnl () ++
-            str "because" ++ spc () ++ str explanation))
+      let ppar = pr_disjunction (fun s -> quote (pr_sort_family s)) sorts in
+      let ppt = pr_lconstr_env ctx (snd (decompose_prod_assum pj.uj_type)) in
+      hov 0 
+	(str "the return type has sort" ++ spc() ++ ppt ++ spc() ++ 
+	 str "while it" ++ spc() ++ str "should be " ++ ppar ++ str ".") ++ 
+      fnl() ++ 
+      hov 0
+	(str "Elimination of an inductive object of sort " ++
+	 pki ++ brk(1,0) ++
+         str "is not allowed on a predicate in sort " ++ pkp ++ fnl() ++
+         str "because" ++ spc() ++ str explanation ++ str ".")
   | None -> 
-      mt ()
+      str "ill-formed elimination predicate."
   in
-  str "Incorrect elimination of" ++ brk(1,1) ++ pc ++ spc () ++
-  str "in the inductive type" ++ brk(1,1) ++ pi ++ fnl () ++
-  str "The elimination predicate" ++ brk(1,1) ++ pp ++ spc () ++
-  str "has type" ++ brk(1,1) ++ ppt ++ fnl () ++
-  str "It should be one of :" ++ brk(1,1)  ++ hov 0 ppar ++ fnl () ++
-  msg
+  hov 0 (
+    str "Incorrect elimination of" ++ spc() ++ pc ++ spc () ++
+    str "in the inductive type" ++ spc() ++ quote pi ++ str":") ++ 
+  fnl () ++ msg
 
 let explain_case_not_inductive ctx cj =
-  let pc = prterm_env ctx cj.uj_val in
-  let pct = prterm_env ctx cj.uj_type in
+  let ctx = make_all_name_different ctx in
+  let pc = pr_lconstr_env ctx cj.uj_val in
+  let pct = pr_lconstr_env ctx cj.uj_type in
     match kind_of_term cj.uj_type with
       | Evar _ -> 
 	  str "Cannot infer a type for this expression"
       | _ ->
-	  str "This term" ++ brk(1,1) ++ pc ++ spc () ++ 
+	  str "The term" ++ brk(1,1) ++ pc ++ spc () ++ 
 	  str "has type" ++ brk(1,1) ++ pct ++ spc () ++ 
 	  str "which is not a (co-)inductive type"
 
 let explain_number_branches ctx cj expn =
-  let pc = prterm_env ctx cj.uj_val in
-  let pct = prterm_env ctx cj.uj_type in
-  str "Cases on term" ++ brk(1,1) ++ pc ++ spc ()  ++
+  let ctx = make_all_name_different ctx in
+  let pc = pr_lconstr_env ctx cj.uj_val in
+  let pct = pr_lconstr_env ctx cj.uj_type in
+  str "Matching on term" ++ brk(1,1) ++ pc ++ spc ()  ++
   str "of type" ++ brk(1,1) ++ pct ++ spc () ++
   str "expects " ++  int expn ++ str " branches"
 
+let ordinal n =
+  let s = match n mod 10 with 1 -> "st" | 2 -> "nd" | 3 -> "rd" | _ -> "th" in
+  string_of_int n ^ s
+
 let explain_ill_formed_branch ctx c i actty expty =
-  let pc = prterm_env ctx c in
-  let pa = prterm_env ctx actty in
-  let pe = prterm_env ctx expty in
-  str "In Cases expression on term" ++ brk(1,1) ++ pc ++
-  spc () ++ str "the branch "  ++ int (i+1) ++
-  str " has type" ++ brk(1,1) ++ pa  ++ spc () ++ 
+  let ctx = make_all_name_different ctx in
+  let pc = pr_lconstr_env ctx c in
+  let pa = pr_lconstr_env ctx actty in
+  let pe = pr_lconstr_env ctx expty in
+  str "In pattern-matching on term" ++ brk(1,1) ++ pc ++
+  spc () ++ str "the " ++ str (ordinal (i+1)) ++ str " branch has type" ++
+  brk(1,1) ++ pa ++ spc () ++ 
   str "which should be" ++ brk(1,1) ++ pe
 
 let explain_generalization ctx (name,var) j =
-  let ctx = make_all_name_different ctx in
   let pe = pr_ne_context_of (str "In environment") ctx in
-  let pv = prtype_env ctx var in
-  let (pc,pt) = prjudge_env (push_rel_assum (name,var) ctx) j in
+  let pv = pr_ltype_env ctx var in
+  let (pc,pt) = pr_ljudge_env (push_rel_assum (name,var) ctx) j in
   str"Illegal generalization: " ++ pe ++
   str"Cannot generalize" ++ brk(1,1) ++ pv ++ spc () ++
   str"over" ++ brk(1,1) ++ pc ++ str"," ++ spc () ++ 
@@ -131,50 +160,48 @@ let explain_generalization ctx (name,var) j =
   spc () ++ str"which should be Set, Prop or Type."
 
 let explain_actual_type ctx j pt =
-  let ctx = make_all_name_different ctx in
   let pe = pr_ne_context_of (str "In environment") ctx in
-  let (pc,pct) = prjudge_env ctx j in
-  let pt = prterm_env ctx pt in
+  let (pc,pct) = pr_ljudge_env ctx j in
+  let pt = pr_lconstr_env ctx pt in
   pe ++
-  str "The term" ++ brk(1,1) ++ pc  ++ spc ()  ++
+  str "The term" ++ brk(1,1) ++ pc ++ spc ()  ++
   str "has type"  ++ brk(1,1) ++ pct ++ brk(1,1) ++ 
   str "while it is expected to have type" ++ brk(1,1) ++ pt
 
 let explain_cant_apply_bad_type ctx (n,exptyp,actualtyp) rator randl =
-  let randl = Array.to_list randl in
   let ctx = make_all_name_different ctx in
+  let randl = Array.to_list randl in
 (*  let pe = pr_ne_context_of (str"in environment") ctx in*)
-  let pr,prt = prjudge_env ctx rator in
+  let pr,prt = pr_ljudge_env ctx rator in
   let term_string1,term_string2 =
     if List.length randl > 1 then
-      let many = match n mod 10 with 1 -> "st" | 2 -> "nd" | _ -> "th" in
-      "terms", "The "^(string_of_int n)^many^" term"
+      str "terms", (str"The "++nth n++str" term")
     else
-      "term","This term" in
+      str "term", str "This term" in
   let appl = prlist_with_sep pr_fnl 
 	       (fun c ->
-		  let pc,pct = prjudge_env ctx c in
+		  let pc,pct = pr_ljudge_env ctx c in
 		  hov 2 (pc ++ spc () ++ str": "  ++ pct)) randl
   in
   str"Illegal application (Type Error): " ++ (* pe ++ *) fnl () ++
   str"The term" ++ brk(1,1) ++ pr ++ spc () ++
   str"of type" ++ brk(1,1) ++ prt ++ spc ()  ++
-  str("cannot be applied to the "^term_string1) ++ fnl () ++ 
-  str" " ++ v 0 appl ++ fnl () ++ str (term_string2^" has type") ++
-  brk(1,1) ++ prterm_env ctx actualtyp ++ spc () ++
-  str"which should be coercible to" ++ brk(1,1) ++ prterm_env ctx exptyp
+  str"cannot be applied to the " ++ term_string1 ++ fnl () ++ 
+  str" " ++ v 0 appl ++ fnl () ++ term_string2 ++ str" has type" ++
+  brk(1,1) ++ pr_lconstr_env ctx actualtyp ++ spc () ++
+  str"which should be coercible to" ++ brk(1,1) ++ pr_lconstr_env ctx exptyp
 
 let explain_cant_apply_not_functional ctx rator randl =
-  let randl = Array.to_list randl in
   let ctx = make_all_name_different ctx in
+  let randl = Array.to_list randl in
 (*  let pe = pr_ne_context_of (str"in environment") ctx in*)
-  let pr = prterm_env ctx rator.uj_val in
-  let prt = prterm_env ctx (body_of_type rator.uj_type) in
+  let pr = pr_lconstr_env ctx rator.uj_val in
+  let prt = pr_lconstr_env ctx rator.uj_type in
   let term_string = if List.length randl > 1 then "terms" else "term" in
   let appl = prlist_with_sep pr_fnl 
 	       (fun c ->
-		  let pc = prterm_env ctx c.uj_val in
-		  let pct = prterm_env ctx (body_of_type c.uj_type) in
+		  let pc = pr_lconstr_env ctx c.uj_val in
+		  let pct = pr_lconstr_env ctx c.uj_type in
 		  hov 2 (pc ++ spc () ++ str": "  ++ pct)) randl
   in
   str"Illegal application (Non-functional construction): " ++ 
@@ -185,117 +212,205 @@ let explain_cant_apply_not_functional ctx rator randl =
   str" " ++ v 0 appl
 
 let explain_unexpected_type ctx actual_type expected_type =
-  let ctx = make_all_name_different ctx in
-  let pract = prterm_env ctx actual_type in
-  let prexp = prterm_env ctx expected_type in
+  let pract = pr_lconstr_env ctx actual_type in
+  let prexp = pr_lconstr_env ctx expected_type in
   str"This type is" ++ spc () ++ pract ++ spc () ++ 
   str "but is expected to be" ++
   spc () ++ prexp
 
 let explain_not_product ctx c =
-  let ctx = make_all_name_different ctx in
-  let pr = prterm_env ctx c in
+  let pr = pr_lconstr_env ctx c in
   str"The type of this term is a product," ++ spc () ++
-  str"but it is casted with type" ++
-  brk(1,1) ++ pr
+  str"while it is expected to be" ++ 
+  if is_Type c then str " a sort" else (brk(1,1) ++ pr)
 
 (* TODO: use the names *)
 (* (co)fixpoints *)
-let explain_ill_formed_rec_body ctx err names i vdefs =
+let explain_ill_formed_rec_body ctx err names i =
+  let prt_name i =
+    match names.(i) with
+        Name id -> str "Recursive definition of " ++ pr_id id
+      | Anonymous -> str"The " ++ nth i ++ str" definition" in
+
   let st = match err with
 
   (* Fixpoint guard errors *)
   | NotEnoughAbstractionInFixBody ->
       str "Not enough abstractions in the definition"
-  | RecursionNotOnInductiveType ->
-      str "Recursive definition on a non inductive type"
-  | RecursionOnIllegalTerm ->
-      str "Recursive call applied to an illegal term"
-  | NotEnoughArgumentsForFixCall ->
-      str "Not enough arguments for the recursive call"
+  | RecursionNotOnInductiveType c ->
+      str "Recursive definition on" ++ spc() ++ pr_lconstr_env ctx c ++ spc() ++
+      str "which should be an inductive type"
+  | RecursionOnIllegalTerm(j,arg,le,lt) ->
+      let called =
+        match names.(j) with
+            Name id -> pr_id id
+          | Anonymous -> str"the " ++ nth i ++ str" definition" in
+      let vars =
+        match (lt,le) with
+            ([],[]) -> mt()
+          | ([],[x]) ->
+              str "a subterm of " ++ pr_db ctx x
+          | ([],_) ->
+              str "a subterm of the following variables: " ++
+              prlist_with_sep pr_spc (pr_db ctx) le
+          | ([x],_) -> pr_db ctx x
+          | _ ->
+              str "one of the following variables: " ++
+              prlist_with_sep pr_spc (pr_db ctx) lt in
+      str "Recursive call to " ++ called ++ spc() ++
+      str "has principal argument equal to" ++ spc() ++
+      pr_lconstr_env ctx arg ++ fnl() ++ str "instead of " ++ vars
+
+  | NotEnoughArgumentsForFixCall j ->
+      let called =
+        match names.(j) with
+            Name id -> pr_id id
+          | Anonymous -> str"the " ++ nth i ++ str" definition" in
+     str "Recursive call to " ++ called ++ str " had not enough arguments"
 
   (* CoFixpoint guard errors *)
-  (* TODO : rcuprer le contexte des termes pour pouvoir les afficher *)
   | CodomainNotInductiveType c ->
-      str "The codomain is" ++ spc () ++ prterm c ++ spc () ++
+      str "the codomain is" ++ spc () ++ pr_lconstr_env ctx c ++ spc () ++
       str "which should be a coinductive type"
   | NestedRecursiveOccurrences ->
-      str "Nested recursive occurrences"
+      str "nested recursive occurrences"
   | UnguardedRecursiveCall c ->
-      str "Unguarded recursive call"
+      str "unguarded recursive call in" ++ spc() ++ pr_lconstr_env ctx c
   | RecCallInTypeOfAbstraction c ->
-      str "Not allowed recursive call in the domain of an abstraction"
+      str "recursive call forbidden in the domain of an abstraction:" ++
+      spc() ++ pr_lconstr_env ctx c
   | RecCallInNonRecArgOfConstructor c ->
-      str "Not allowed recursive call in a non-recursive argument of constructor"
+      str "recursive call on a non-recursive argument of constructor" ++
+      spc() ++ pr_lconstr_env ctx c
   | RecCallInTypeOfDef c ->
-      str "Not allowed recursive call in the type of a recursive definition"
+      str "recursive call forbidden in the type of a recursive definition" ++
+      spc() ++ pr_lconstr_env ctx c
   | RecCallInCaseFun c ->
-      str "Not allowed recursive call in a branch of cases"
+      str "recursive call in a branch of" ++ spc() ++ pr_lconstr_env ctx c
   | RecCallInCaseArg c -> 
-      str "Not allowed recursive call in the argument of cases"
+      str "recursive call in the argument of cases in" ++ spc() ++
+      pr_lconstr_env ctx c
   | RecCallInCasePred c ->
-      str "Not allowed recursive call in the type of cases in"
-  | NotGuardedForm ->
-      str "Definition not in guarded form"
+      str "recursive call in the type of cases in" ++  spc() ++
+      pr_lconstr_env ctx c
+  | NotGuardedForm c ->
+      str "sub-expression " ++ pr_lconstr_env ctx c ++ spc() ++
+      str "not in guarded form" ++ spc()++
+      str"(should be a constructor, an abstraction, a match, a cofix or a recursive call)"
   in
-  let pvd = prterm_env ctx vdefs.(i) in
-  let s = match names.(i) with Name id -> string_of_id id | Anonymous -> "_" in
-  st ++ fnl () ++ str"The " ++
-  (if Array.length vdefs = 1 then mt () else (int (i+1) ++ str "-th ")) ++
-  str"recursive definition" ++ spc () ++ str s ++
-  spc ()  ++ str":=" ++ spc ()  ++ pvd ++ spc () ++
-  str "is not well-formed"
+  prt_name i ++ str" is ill-formed." ++ fnl() ++
+  pr_ne_context_of (str "In environment") ctx ++
+  st
 
 let explain_ill_typed_rec_body ctx i names vdefj vargs =
-  let pvd,pvdt = prjudge_env ctx (vdefj.(i)) in
-  let pv = prterm_env ctx (body_of_type vargs.(i)) in
+  let ctx = make_all_name_different ctx in
+  let pvd,pvdt = pr_ljudge_env ctx (vdefj.(i)) in
+  let pv = pr_lconstr_env ctx vargs.(i) in
   str"The "  ++
   (if Array.length vdefj = 1 then mt () else int (i+1) ++ str "-th") ++
   str"recursive definition"  ++ spc () ++ pvd ++ spc () ++
   str "has type" ++ spc () ++ pvdt ++spc () ++ 
   str "it should be" ++ spc () ++ pv 
-
+(*
 let explain_not_inductive ctx c =
-  let pc = prterm_env ctx c in
+  let ctx = make_all_name_different ctx in
+  let pc = pr_lconstr_env ctx c in
   str"The term" ++ brk(1,1) ++ pc ++ spc () ++
   str "is not an inductive definition"
-
+*)
 let explain_cant_find_case_type ctx c =
-  let pe = prterm_env ctx c in
+  let ctx = make_all_name_different ctx in
+  let pe = pr_lconstr_env ctx c in
   hov 3 (str "Cannot infer type of pattern-matching on" ++ ws 1 ++ pe)
 
 let explain_occur_check ctx ev rhs =
-  let id = "?" ^ string_of_int ev in
-  let pt = prterm_env ctx rhs in
+  let ctx = make_all_name_different ctx in
+  let id = Evd.string_of_existential ev in
+  let pt = pr_lconstr_env ctx rhs in
   str"Occur check failed: tried to define " ++ str id ++
   str" with term" ++ brk(1,1) ++ pt
 
-let explain_not_clean ctx ev t =
+let explain_hole_kind env = function
+  | QuestionMark -> str "a term for this placeholder"
+  | CasesType ->
+      str "the type of this pattern-matching problem"
+  | BinderType (Name id) ->
+      str "a type for " ++ Nameops.pr_id id
+  | BinderType Anonymous ->
+      str "a type for this anonymous binder"
+  | ImplicitArg (c,(n,ido)) ->
+      let id = out_some ido in
+      str "an instance for the implicit parameter " ++
+      pr_id id ++ spc () ++ str "of" ++
+      spc () ++ Nametab.pr_global_env Idset.empty c
+  | InternalHole ->
+      str "a term for an internal placeholder"
+  | TomatchTypeParameter (tyi,n) ->
+      str "the " ++ pr_ord n ++
+      str " argument of the inductive type (" ++ pr_inductive env tyi ++
+      str ") of this term"
+
+let explain_not_clean ctx ev t k =
+  let ctx = make_all_name_different ctx in
   let c = mkRel (Intset.choose (free_rels t)) in
-  let id = "?" ^ string_of_int ev in
-  let var = prterm_env ctx c in
-  str"Tried to define " ++ str id ++
-  str" with a term using variable " ++ var ++ spc () ++
+  let id = Evd.string_of_existential ev in
+  let var = pr_lconstr_env ctx c in
+  str"Tried to define " ++ explain_hole_kind ctx k ++
+  str" (" ++ str id ++ str ")" ++ spc() ++
+  str"with a term using variable " ++ var ++ spc () ++
   str"which is not in its scope."
 
+let explain_unsolvable_implicit env k =
+  str "Cannot infer " ++ explain_hole_kind env k
+
+
 let explain_var_not_found ctx id = 
   str "The variable" ++ spc () ++ str (string_of_id id) ++
   spc ()  ++ str "was not found" ++ 
   spc ()  ++ str "in the current" ++ spc ()  ++ str "environment"
 
 let explain_wrong_case_info ctx ind ci =
-  let pi = prterm (mkInd ind) in
+  let pi = pr_lconstr (mkInd ind) in
   if ci.ci_ind = ind then
-    str"Cases expression on an object of inductive" ++ spc () ++ pi ++
+    str"Pattern-matching expression on an object of inductive" ++ spc () ++ pi ++
     spc () ++ str"has invalid information"
   else
-    let pc = prterm (mkInd ci.ci_ind) in
+    let pc = pr_lconstr (mkInd ci.ci_ind) in
     str"A term of inductive type" ++ spc () ++ pi ++ spc () ++
-    str"was given to a Cases expression on the inductive type" ++
+    str"was given to a pattern-matching expression on the inductive type" ++
     spc () ++ pc
        
 
-let explain_type_error ctx = function
+let explain_cannot_unify m n =
+  let pm = pr_lconstr m in 
+  let pn = pr_lconstr n in
+  str"Impossible to unify" ++ brk(1,1)  ++ pm ++ spc ()  ++
+  str"with" ++ brk(1,1)  ++ pn
+
+let explain_cannot_unify_local env m n subn =
+  let pm = pr_lconstr m in 
+  let pn = pr_lconstr n in
+  let psubn = pr_lconstr_env env subn in
+    str"Impossible to unify" ++ brk(1,1)  ++ pm ++ spc ()  ++
+      str"with" ++ brk(1,1)  ++ pn ++ spc() ++ str"as" ++ brk(1,1) ++ 
+      psubn ++ str" contains local variables"
+
+let explain_refiner_cannot_generalize ty =
+  str "Cannot find a well-typed generalisation of the goal with type : " ++ 
+  pr_lconstr ty
+
+let explain_no_occurrence_found c =
+  str "Found no subterm matching " ++ pr_lconstr c ++ str " in the current goal"
+
+let explain_cannot_unify_binding_type m n =
+  let pm = pr_lconstr m in 
+  let pn = pr_lconstr n in
+  str "This binding has type" ++ brk(1,1) ++ pm ++ spc () ++ 
+  str "which should be unifiable with" ++ brk(1,1) ++ pn
+
+let explain_type_error ctx err =
+  let ctx = make_all_name_different ctx in
+  match err with
   | UnboundRel n -> 
       explain_unbound_rel ctx n
   | UnboundVar v -> 
@@ -322,8 +437,8 @@ let explain_type_error ctx = function
       explain_cant_apply_bad_type ctx t rator randl
   | CantApplyNonFunctional (rator, randl) ->
       explain_cant_apply_not_functional ctx rator randl
-  | IllFormedRecBody (i, lna, vdefj, vargs) ->
-      explain_ill_formed_rec_body ctx i lna vdefj vargs
+  | IllFormedRecBody (err, lna, i) ->
+      explain_ill_formed_rec_body ctx err lna i
   | IllTypedRecBody (i, lna, vdefj, vargs) -> 
      explain_ill_typed_rec_body ctx i lna vdefj vargs
   | WrongCaseInfo (ind,ci) ->
@@ -332,79 +447,71 @@ let explain_type_error ctx = function
   | NotInductive c ->
       explain_not_inductive ctx c
 *)
-let explain_pretype_error ctx = function
+let explain_pretype_error ctx err =
+  let ctx = make_all_name_different ctx in
+  match err with
   | CantFindCaseType c ->
       explain_cant_find_case_type ctx c
   | OccurCheck (n,c) ->
       explain_occur_check ctx n c
-  | NotClean (n,c) ->
-      explain_not_clean ctx n c
+  | NotClean (n,c,k) ->
+      explain_not_clean ctx n c k
+  | UnsolvableImplicit k ->
+      explain_unsolvable_implicit ctx k
   | VarNotFound id ->
       explain_var_not_found ctx id
   | UnexpectedType (actual,expected) ->
       explain_unexpected_type ctx actual expected
   | NotProduct c ->
       explain_not_product ctx c
+  | CannotUnify (m,n) -> explain_cannot_unify m n
+  | CannotUnifyLocal (e,m,n,sn) -> explain_cannot_unify_local e m n sn
+  | CannotGeneralize ty -> explain_refiner_cannot_generalize ty
+  | NoOccurrenceFound c -> explain_no_occurrence_found c
+  | CannotUnifyBindingType (m,n) -> explain_cannot_unify_binding_type m n
 
 (* Refiner errors *)
 
 let explain_refiner_bad_type arg ty conclty =
   str"refiner was given an argument" ++ brk(1,1) ++ 
-  prterm arg ++ spc () ++
-  str"of type" ++ brk(1,1) ++ prterm ty ++ spc () ++
-  str"instead of" ++ brk(1,1) ++ prterm conclty
+  pr_lconstr arg ++ spc () ++
+  str"of type" ++ brk(1,1) ++ pr_lconstr ty ++ spc () ++
+  str"instead of" ++ brk(1,1) ++ pr_lconstr conclty
 
 let explain_refiner_occur_meta t =
-  str"cannot refine with term" ++ brk(1,1) ++ prterm t ++
+  str"cannot refine with term" ++ brk(1,1) ++ pr_lconstr t ++
   spc () ++ str"because there are metavariables, and it is" ++
   spc () ++ str"neither an application nor a Case"
 
 let explain_refiner_occur_meta_goal t =
-  str"generated subgoal" ++ brk(1,1) ++ prterm t ++
+  str"generated subgoal" ++ brk(1,1) ++ pr_lconstr t ++
   spc () ++ str"has metavariables in it"
 
-let explain_refiner_cannot_applt t harg =
+let explain_refiner_cannot_apply t harg =
   str"in refiner, a term of type " ++ brk(1,1) ++
-  prterm t ++ spc () ++ str"could not be applied to" ++ brk(1,1) ++
-  prterm harg
-
-let explain_refiner_cannot_unify m n =
-  let pm = prterm m in 
-  let pn = prterm n in
-  str"Impossible to unify" ++ brk(1,1)  ++ pm ++ spc ()  ++
-  str"with" ++ brk(1,1)  ++ pn
-
-let explain_refiner_cannot_generalize ty =
-  str "Cannot find a well-typed generalisation of the goal with type : " ++ 
-  prterm ty
+  pr_lconstr t ++ spc () ++ str"could not be applied to" ++ brk(1,1) ++
+  pr_lconstr harg
 
 let explain_refiner_not_well_typed c =
-  str"The term "  ++ prterm c  ++ str" is not well-typed"
-
-let explain_refiner_bad_tactic_args s l =
-  str "Internal tactic " ++ str s ++ str " cannot be applied to " ++
-  Tacmach.pr_tactic (s,l)
+  str"The term "  ++ pr_lconstr c  ++ str" is not well-typed"
 
 let explain_intro_needs_product () =
   str "Introduction tactics needs products"
 
 let explain_does_not_occur_in c hyp =
-  str "The term" ++ spc () ++ prterm c ++ spc () ++ str "does not occur in" ++
+  str "The term" ++ spc () ++ pr_lconstr c ++ spc () ++ str "does not occur in" ++ 
   spc () ++ pr_id hyp
 
 let explain_non_linear_proof c =
-  str "cannot refine with term" ++ brk(1,1) ++ prterm c ++
+  str "cannot refine with term" ++ brk(1,1) ++ pr_lconstr c ++
   spc () ++ str"because a metavariable has several occurrences"
 
 let explain_refiner_error = function
   | BadType (arg,ty,conclty) -> explain_refiner_bad_type arg ty conclty
   | OccurMeta t -> explain_refiner_occur_meta t
   | OccurMetaGoal t -> explain_refiner_occur_meta_goal t
-  | CannotApply (t,harg) -> explain_refiner_cannot_applt t harg
-  | CannotUnify (m,n) -> explain_refiner_cannot_unify m n
-  | CannotGeneralize ty -> explain_refiner_cannot_generalize ty
+  | CannotApply (t,harg) -> explain_refiner_cannot_apply t harg
   | NotWellTyped c -> explain_refiner_not_well_typed c
-  | BadTacticArgs (s,l) -> explain_refiner_bad_tactic_args s l
   | IntroNeedsProduct -> explain_intro_needs_product ()
   | DoesNotOccurIn (c,hyp) -> explain_does_not_occur_in c hyp
   | NonLinearProof c -> explain_non_linear_proof c
@@ -412,20 +519,20 @@ let explain_refiner_error = function
 (* Inductive errors *)
 
 let error_non_strictly_positive env c v  =
-  let pc = prterm_env env c in
-  let pv = prterm_env env v in
+  let pc = pr_lconstr_env env c in
+  let pv = pr_lconstr_env env v in
   str "Non strictly positive occurrence of " ++ pv ++ str " in" ++
   brk(1,1) ++ pc
 
 let error_ill_formed_inductive env c v =
-  let pc = prterm_env env c in
-  let pv = prterm_env env v in
+  let pc = pr_lconstr_env env c in
+  let pv = pr_lconstr_env env v in
   str "Not enough arguments applied to the " ++ pv ++
   str " in" ++ brk(1,1) ++ pc
 
 let error_ill_formed_constructor env c v =
-  let pc = prterm_env env c in
-  let pv = prterm_env env v in
+  let pc = pr_lconstr_env env c in
+  let pv = pr_lconstr_env env v in
   str "The conclusion of" ++ brk(1,1) ++ pc ++ brk(1,1) ++ 
   str "is not valid;" ++ brk(1,1) ++ str "it must be built from " ++ pv
 
@@ -438,9 +545,9 @@ let str_of_nth n =
      | _ -> "th")
 
 let error_bad_ind_parameters env c n v1 v2  =
-  let pc = prterm_env_at_top env c in
-  let pv1 = prterm_env env v1 in
-  let pv2 = prterm_env env v2 in
+  let pc = pr_lconstr_env_at_top env c in
+  let pv1 = pr_lconstr_env env v1 in
+  let pv2 = pr_lconstr_env env v2 in
   str ("The "^(str_of_nth n)^" argument of ") ++ pv2 ++ brk(1,1) ++
   str "must be " ++ pv1 ++ str " in" ++ brk(1,1) ++ pc
 
@@ -453,6 +560,11 @@ let error_same_names_constructors id cid =
   str "is used twice is the definition of type" ++ spc () ++
   pr_id id
 
+let error_same_names_overlap idl = 
+  str "The following names" ++ spc () ++ 
+  str "are used both as type names and constructor names:" ++ spc () ++
+  prlist_with_sep pr_coma pr_id idl
+
 let error_not_an_arity id =
   str "The type of" ++ spc () ++ pr_id id ++ spc () ++ str "is not an arity."
 
@@ -461,30 +573,35 @@ let error_bad_entry () =
 
 let error_not_allowed_case_analysis dep kind i =
   str (if dep then "Dependent" else "Non Dependent") ++
-  str " case analysis on sort: " ++ print_sort kind ++ fnl () ++
+  str " case analysis on sort: " ++ pr_sort kind ++ fnl () ++
   str "is not allowed for inductive definition: " ++
   pr_inductive (Global.env()) i
 
 let error_bad_induction dep indid kind =
   str (if dep then "Dependent" else "Non dependent") ++
   str " induction for type " ++ pr_id indid ++
-  str " and sort " ++ print_sort kind ++ spc () ++
+  str " and sort " ++ pr_sort kind ++ spc () ++
   str "is not allowed"
 
 let error_not_mutual_in_scheme () =
- str "Induction schemes is concerned only with mutually inductive types"
+ str "Induction schemes are concerned only with distinct mutually inductive types"
+
+(* Inductive constructions errors *)
 
 let explain_inductive_error = function
-  (* These are errors related to inductive constructions *)
   | NonPos (env,c,v) -> error_non_strictly_positive env c v
   | NotEnoughArgs (env,c,v) -> error_ill_formed_inductive env c v
   | NotConstructor (env,c,v) -> error_ill_formed_constructor env c v
   | NonPar (env,c,n,v1,v2) -> error_bad_ind_parameters env c n v1 v2
   | SameNamesTypes id -> error_same_names_types id
   | SameNamesConstructors (id,cid) -> error_same_names_constructors id cid
+  | SameNamesOverlap idl -> error_same_names_overlap idl
   | NotAnArity id -> error_not_an_arity id
   | BadEntry -> error_bad_entry ()
-  (* These are errors related to recursors *)
+
+(* Recursion schemes errors *)
+
+let explain_recursion_scheme_error = function
   | NotAllowedCaseAnalysis (dep,k,i) ->
       error_not_allowed_case_analysis dep k i
   | BadInduction (dep,indid,kind) -> error_bad_induction dep indid kind
@@ -493,7 +610,8 @@ let explain_inductive_error = function
 (* Pattern-matching errors *)
 
 let explain_bad_pattern ctx cstr ty = 
-  let pt = prterm_env ctx ty in
+  let ctx = make_all_name_different ctx in
+  let pt = pr_lconstr_env ctx ty in
   let pc = pr_constructor ctx cstr in
   str "Found the constructor " ++ pc ++ brk(1,1) ++ 
   str "while matching a term of type " ++ pt ++ brk(1,1) ++ 
@@ -507,29 +625,38 @@ let explain_bad_constructor ctx cstr ind =
   str "while a constructor of "  ++ pi ++ brk(1,1)  ++
   str "is expected"
 
-let explain_wrong_numarg_of_constructor ctx cstr n =
-  let pc = pr_constructor ctx cstr in
-  str "The constructor " ++ pc ++ str " expects "  ++
-  (if n = 0 then str "no argument." else if n = 1 then str "1 argument."
-   else (int n  ++ str " arguments."))
+let decline_string n s =
+  if n = 0 then "no "^s
+  else if n = 1 then "1 "^s
+  else (string_of_int n^" "^s^"s")
+
+let explain_wrong_numarg_constructor ctx cstr n =
+  str "The constructor " ++ pr_constructor ctx cstr ++ 
+  str " expects "  ++ str (decline_string n "argument")
+
+let explain_wrong_numarg_inductive ctx ind n =
+  str "The inductive type " ++ pr_inductive ctx ind ++ 
+  str " expects "  ++ str (decline_string n "argument")
 
 let explain_wrong_predicate_arity ctx pred nondep_arity dep_arity=
-  let pp = prterm_env ctx pred in
+  let ctx = make_all_name_different ctx in
+  let pp = pr_lconstr_env ctx pred in
   str "The elimination predicate " ++ spc () ++ pp ++ fnl () ++
   str "should be of arity"  ++ spc () ++
-  prterm_env ctx nondep_arity  ++ spc () ++ 
+  pr_lconstr_env ctx nondep_arity  ++ spc () ++ 
   str "(for non dependent case) or"  ++
-  spc () ++ prterm_env ctx dep_arity  ++ spc () ++ str "(for dependent case)."
+  spc () ++ pr_lconstr_env ctx dep_arity  ++ spc () ++ str "(for dependent case)."
 
 let explain_needs_inversion ctx x t =
-  let px = prterm_env ctx x in
-  let pt = prterm_env ctx t in
+  let ctx = make_all_name_different ctx in
+  let px = pr_lconstr_env ctx x in
+  let pt = pr_lconstr_env ctx t in
   str "Sorry, I need inversion to compile pattern matching of term " ++
   px  ++ str " of type: " ++ pt
 
 let explain_unused_clause env pats =
-  let s = if List.length pats > 1 then "s" else "" in
 (* Without localisation
+  let s = if List.length pats > 1 then "s" else "" in
   (str ("Unused clause with pattern"^s) ++ spc () ++
     hov 0 (prlist_with_sep pr_spc pr_cases_pattern pats) ++ str ")")
 *)
@@ -540,10 +667,11 @@ let explain_non_exhaustive env pats =
   str ("Non exhaustive pattern-matching: no clause found for pattern"^s) ++
   spc () ++ hov 0 (prlist_with_sep pr_spc pr_cases_pattern pats)
 
-let explain_cannot_infer_predicate env typs =
+let explain_cannot_infer_predicate ctx typs =
+  let ctx = make_all_name_different ctx in
   let pr_branch (cstr,typ) =
     let cstr,_ = decompose_app cstr in
-    str "For " ++ prterm_env env cstr ++ str " : " ++ prterm_env env typ
+    str "For " ++ pr_lconstr_env ctx cstr ++ str " : " ++ pr_lconstr_env ctx typ
   in
   str "Unable to unify the types found in the branches:" ++
   spc () ++ hov 0 (prlist_with_sep pr_fnl pr_branch (Array.to_list typs))
@@ -554,7 +682,9 @@ let explain_pattern_matching_error env = function
   | BadConstructor (c,ind) ->
       explain_bad_constructor env c ind
   | WrongNumargConstructor (c,n) ->
-      explain_wrong_numarg_of_constructor env c n
+      explain_wrong_numarg_constructor env c n
+  | WrongNumargInductive (c,n) ->
+      explain_wrong_numarg_inductive env c n
   | WrongPredicateArity (pred,n,dep) ->
       explain_wrong_predicate_arity env pred n dep
   | NeedsInversion (x,t) ->
@@ -565,3 +695,9 @@ let explain_pattern_matching_error env = function
       explain_non_exhaustive env tms
   | CannotInferPredicate typs ->
       explain_cannot_infer_predicate env typs
+
+let explain_reduction_tactic_error = function
+  | Tacred.InvalidAbstraction (env,c,(env',e)) ->
+      str "The abstracted term" ++ spc() ++ pr_lconstr_env_at_top env c ++ 
+      spc() ++ str "is not well typed." ++ fnl () ++
+      explain_type_error env' e
